<?php
/**
 * Dynamic Delivery Promise Amazon integration.
 */
defined( 'ABSPATH' ) || exit;

if ( ! class_exists( 'NS_MCF_DDP' ) ) {

	/**
	 * DDP class. This is to handle all integration for DDP.
	 */
	class NS_MCF_DDP extends NS_MCF_Integration {


		/**
		 * Initialize class.
		 */
		public function init() {

			add_action( 'admin_notices', array( $this, 'ddp_notice' ) );

			if ( ! $this->is_enabled() ) {
				return;
			}

			add_action( 'woocommerce_after_add_to_cart_form', array( $this, 'add_ddp_form' ), 999 );

			// add_filter( 'woocommerce_cart_item_name', array( $this, 'add_ddp_summary' ), 10, 3 );

			add_action( 'wp_enqueue_scripts', array( $this, 'load_scripts_and_styles' ) );

			add_action( 'wp_ajax_ns_fba_ddp_check', array( $this, 'ajax_ddp_check' ) );
			add_action( 'wp_ajax_nopriv_ns_fba_ddp_check', array( $this, 'ajax_ddp_check' ) );
		}

		/**
		 * Check if the current region is supported for the integration and is enabled.
		 * Currently only allowed in US locations.
		 *
		 * @return boolean
		 */
		public function is_enabled() {
			$is_marketplace_supported = $this->ns_fba->utils->is_ddp_marketplace_supported();
			return $is_marketplace_supported && $this->ns_fba->utils->isset_on( $this->ns_fba->get_option( 'ns_fba_enable_ddp', 'no' ) );
		}

		/**
		 * Zipcode country initials based on region.
		 *
		 * @return string
		 */
		public function show_zipcode_country() {
			$marketplace = $this->ns_fba->get_option( 'ns_fba_marketplace_id' );
			if ( 'A1F83G8C2ARO7P' === $marketplace ) {
				return 'UK';
			}
			return 'US';
		}

		/**
		 * Get the default zipcode based on location.
		 *
		 * @return string
		 */
		public function default_zip_code() {
			$marketplace = $this->ns_fba->get_option( 'ns_fba_marketplace_id' );
			if ( 'A1F83G8C2ARO7P' === $marketplace ) {
				return 'W1D7DH';
			}
			return '90011'; // Default for US.
		}

		/**
		 * Add the DDP notice.
		 *
		 * @return void
		 */
		public function ddp_notice() {
			// Dont show notice if already dismissed.
			if ( ! NS_MCF_Notification::should_show_notification() ) {
				return;
			}
			$url = admin_url( 'admin.php?page=wc-settings&tab=integration&section=fba&dismiss_ddp_notice=true' );
			?>
			<div class="notice notice-info is-dismissible">
				<p>
					<?php esc_html_e( 'Introducing MCF fast badges from Amazon. This offers MCF merchants the opportunity to show delivery time-based estimates per product to their customers.', 'ns-fba-for-woocommerce' ); ?>
					<a href="<?php printf( esc_html( $url ) ); ?>"><?php esc_html_e( 'View settings', 'ns-fba-for-woocommerce' ); ?></a>
				</p>
			</div>
			<?php
		}

		/**
		 * Check the delivery offerings
		 *
		 * @param string $sku      The product SKU.
		 * @param string $zip      The current zip.
		 * @param int    $quantity The quantity.
		 *
		 * @return array
		 */
		public function check_offerings( $sku, $zip, $quantity, $retry = 0 ) {

			$cache_string = sanitize_title( 'ns_fba_ddp_' . $sku . '_' . $zip . '_' . $quantity );
			$offerings    = get_transient( $cache_string );

			if ( false === $offerings ) {
				$path = '/fba/outbound/2020-07-01/deliveryOffers';
				$body = array(
					'product' => array(
						'productIdentifier' => array(
							'merchantSku' => $sku,
						),
						'amount'            => array(
							'unitOfMeasure' => 'Eaches',
							'value'         => $quantity,
						),
					),

					'terms'   => array(
						'origin'      => array(
							'countryCode' => 'US',
						),
						'destination' => array(
							'deliveryAddress' => array(
								'countryCode'      => 'US',
								'postalCode'       => $zip,
								'city'             => '',
								'districtOrCounty' => '',
								'stateOrRegion'    => '',
							),
							'ipAddress'       => '',
						),
					),
				);

				$body     = wp_json_encode( $body );
				$response = $this->ns_fba->sp_api->make_request( $path, 'deliveryOffers', 0, 'POST', $body );

				if ( SP_API::is_error_in( $response ) ) {
					++$retry;
					if ( 3 === $retry ) {
						return array(
							'status'  => false,
							'message' => apply_filters( 'ns_fba_ddp_general_error', __( 'Error getting the estimate. Please try again in a few minutes', 'ns-fba-for-woocommerce' ) ),
						);
					}
					return $this->check_offerings( $sku, $zip, $quantity, $retry ); // Retry
				}

				$this->ns_fba->logger->add_entry( $response, 'wc', '_ddp' );

				$payload = json_decode( $response['body'], true );
				if ( ! is_array( $payload ) || ! isset( $payload['payload'] ) ) {
					return array(
						'status'  => false,
						'message' => apply_filters( 'ns_fba_ddp_general_error', __( 'Error getting the estimate. Please try again in a few minutes', 'ns-fba-for-woocommerce' ) ),
					);
				}

				$offerings = $payload['payload']['deliveryOffers'];

				set_transient( $cache_string, $offerings, HOUR_IN_SECONDS );
			}

			if ( ! is_array( $offerings ) || empty( $offerings ) ) {
				delete_transient( $cache_string );
				return array(
					'status'  => false,
					'message' => apply_filters( 'ns_fba_ddp_general_error', __( 'Error getting the estimate. Please try again in a few minutes', 'ns-fba-for-woocommerce' ) ),
				);
			}

			$message = array();

			foreach ( $offerings as $offer ) {
				$message['text']     = $offer['policy']['messaging']['text'];
				$message['time']     = sprintf( __( 'Earliest %1$s. Latest %2$s', 'ns-fba-for-woocommerce' ), gmdate( 'm-d-Y', strtotime( $offer['dateRange']['earliest'] ) ), gmdate( 'm-d-Y', strtotime( $offer['dateRange']['latest'] ) ) );
				$message['end_text'] = __( 'Order within the next', 'ns-fba-for-woocommerce' );
				$message['end']      = strtotime( $offer['expiresAt'] );
				$message['zip_text'] = __( 'Deliver to', 'ns-fba-for-woocommerce' );
				$message['zip']      = $zip;
			}

			return array(
				'status'  => true,
				'message' => $message,
			);
		}

		/**
		 * Add the DDP form in the frontend for estimates.
		 *
		 * @return void
		 */
		public function add_ddp_form() {
			global $product;

			$is_product_amazon_fulfill = $this->ns_fba->utils->is_product_amazon_fulfill( $product );
			if ( ! $is_product_amazon_fulfill ) {
				return;
			}

			$product_offering = false;
			$postal_code      = false;
			if ( is_user_logged_in() ) {
				$customer = new WC_Customer( get_current_user_id() );
				if ( 0 !== $customer->get_id() ) {
					$postal_code = $customer->get_shipping_postcode() ? $customer->get_shipping_postcode() : $customer->get_billing_postcode();
				}
			}

			if ( ! $postal_code ) {
				$postal_code = $this->get_guest_zipcode(); // Get the guest zip code.
			}

			if ( $postal_code && ! empty( $product->get_sku() ) ) {
				$offering = $this->check_offerings( $product->get_sku(), $postal_code, 1 );

				if ( $offering['status'] ) {
					$product_offering = $offering['message'];
				}
			}
			$additional_class = 'ddp-product-zip';
			if ( $product_offering ) {
				$additional_class .= ' ns-fba-hidden ddp-loaded';
			}
			if ( $postal_code ) {
				$additional_class .= ' has-postal';
			}
			?>
			<div class="fba-ddp ddp ddp-<?php echo esc_attr( $product->get_sku() ); ?>" data-sku="<?php echo esc_attr( $product->get_sku() ); ?>" data-zip="<?php echo esc_attr( $postal_code ); ?>">
				<p class="ddp-product-offering <?php echo ( ! $product_offering ) ? 'ns-fba-hidden' : ''; ?>">
					<?php if ( $product_offering ) { ?>
						<span><b><?php echo esc_html( $product_offering['text'] ); ?></b></span><br/>
						<span class="ns-mcf-fba-ddp-information ns-fba-hidden"><?php echo esc_html( $product_offering['end_text'] ); ?></span> <span class="countdown ns-mcf-fba-ddp-time" data-time="<?php echo esc_attr( $product_offering['end'] ); ?>"></span><br>
						<span class="ns-mcf-ddp-delivery-details">
							<?php echo esc_html( $product_offering['zip_text'] ); ?>
							<span class="underlined ns-mcf-fba-ddp-edit-link" data-sku="<?php echo esc_attr( $product->get_sku() ); ?>"><?php echo esc_html( $product_offering['zip'] ); ?></span>
						</span>
					<?php } ?>
				</p>
				<div class="<?php echo esc_attr( $additional_class ); ?>">
					<h4><?php esc_html_e( 'Enter zip code to calculate shipping time', 'ns-fba-for-woocommerce' ); ?></h4>
					<div class="input-wrapper">
						<input type="text" class="input-text ns-fba-ddp-lookup-postal" value="<?php echo esc_attr( $postal_code ); ?>" placeholder="<?php printf( esc_attr__( 'Enter a %s zip code', 'ns-fba-for-woocommerce' ), esc_attr( $this->show_zipcode_country() ) ); ?>"/>
						<button class="button alt ns-fba-ddp-lookup vc_general" data-sku="<?php echo esc_attr( $product->get_sku() ); ?>"><?php esc_html_e( 'Apply', 'ns-fba-for-woocommerce' ); ?></button>
					</div>
					<span class="ddp-response-message error"></span>
				</div>
				<div class="ns-mcf-fba-preloader">
					<div class="loader-bar"></div>
					<div class="loader-bar"></div>
					<div class="loader-bar"></div>
					<div class="filler"></div>
				</div>
			</div>
			<?php
		}

		/**
		 * Add summary on checkout page for the shipping times.
		 *
		 * @return string
		 */
		public function add_ddp_summary( $product_name, $item, $item_id ) {
			$product = $item['data'];

			$is_product_amazon_fulfill = $this->ns_fba->utils->is_product_amazon_fulfill( $product );
			if ( ! $is_product_amazon_fulfill ) {
				return $product_name;
			}

			$customer = new WC_Customer( get_current_user_id() );
			if ( 0 !== $customer->get_id() ) {
				$postal_code = $customer->get_shipping_postcode() ? $customer->get_shipping_postcode() : $customer->get_billing_postcode();
				$offering    = $this->check_offerings( $product->get_sku(), $postal_code, 1 );
				if ( $offering['status'] ) {
					$product_offering = $offering['message'];
					$product_name    .= "<br/><span class='ns-fba-ddp-product-delivery-summpry' style='font-size: .875em;'>(" . esc_html( $product_offering['text'] ) . ')</span><br/>';
				}
			}

			return $product_name;
		}

		/**
		 * Load scripts used for the frontend.
		 *
		 * @return void
		 */
		public function load_scripts_and_styles() {
			// Check if its the product page. If not, do not load.
			if ( ! is_product() ) {
				return;
			}

			wp_enqueue_script(
				'ns-fba-product',
				$this->ns_fba->plugin_url . 'js/ns-fba-product.js',
				array( 'jquery' ),
				$this->ns_fba->version,
				true
			);

			wp_localize_script(
				'ns-fba-product',
				'ns_fba_product',
				array(
					'nonce'     => wp_create_nonce( 'ns-fba-ajax' ),
					'spinner'   => esc_url( get_admin_url() . 'images/spinner.gif' ),
					'ajax_url'  => admin_url( 'admin-ajax.php' ),
					'edit'      => __( 'Edit', 'ns-fba-for-woocommerce' ),
					'error'     => __( 'An error occured', 'ns-fba-for-woocommerce' ),
					'zip_error' => apply_filters( 'ns_fba_ddp_product_zipcode_error', sprintf( __( 'Please enter a valid %s zip code', 'ns-fba-for-woocommerce' ), $this->show_zipcode_country() ), $this->show_zipcode_country() ),
				)
			);

			wp_enqueue_style(
				'ns-fba-product',
				$this->ns_fba->plugin_url . 'css/ns-fba-product.css'
			);
		}

		/**
		 * Ajax action to perfrom api check.
		 * REturn json based on the response.
		 *
		 * @return void
		 */
		public function ajax_ddp_check() {
			$nonce = isset( $_POST['_wpnonce'] ) ? sanitize_text_field( wp_unslash( $_POST['_wpnonce'] ) ) : false;
			if ( ! $nonce || ! wp_verify_nonce( $nonce, 'ns-fba-ajax' ) ) {
				wp_send_json_error( __( 'Invalid request', 'ns-fba-for-woocommerce' ) );
			}

			$sku       = isset( $_POST['sku'] ) ? sanitize_text_field( wp_unslash( $_POST['sku'] ) ) : '';
			$zipcode   = isset( $_POST['zipcode'] ) ? sanitize_text_field( wp_unslash( $_POST['zipcode'] ) ) : '';
			$quantity  = isset( $_POST['quantity'] ) ? (int) sanitize_text_field( wp_unslash( $_POST['quantity'] ) ) : 1;
			$variation = isset( $_POST['variation'] ) ? (int) sanitize_text_field( wp_unslash( $_POST['variation'] ) ) : false;

			if ( empty( $sku ) && $variation ) {
				$product = wc_get_product( $variation );
				$sku     = $this->ns_fba->utils->get_sku_to_send( $product );
			}

			if ( empty( $sku ) ) {
				wp_send_json_error( apply_filters( 'ns_fba_ddp_product_sku_error', __( 'Please select a product with a valid SKU', 'ns-fba-for-woocommerce' ) ) );
			}

			if ( ! preg_match( '/^\d{5}$/', $zipcode ) ) {
				wp_send_json_error( apply_filters( 'ns_fba_ddp_product_zipcode_error', sprintf( __( 'Please enter a valid %s zip code', 'ns-fba-for-woocommerce' ), $this->show_zipcode_country() ), $this->show_zipcode_country() ) );
			}

			$offering = $this->check_offerings( $sku, $zipcode, $quantity );

			wp_send_json_success( $offering );
		}

		/**
		 * Get the zipcode from the IP address
		 *
		 * @return string
		 */
		public function get_guest_zipcode() {
			$ip_address = NS_MCF_IP::get_user_ip();

			// Localhost cannot be checked.
			if ( $ip_address === '127.0.0.1' ) {
				return $this->default_zip_code();
			}

			$response = wp_remote_get( SP_API_ZIP_LOOKUP_URL . '?ip_address=' . $ip_address );

			if ( 200 !== wp_remote_retrieve_response_code( $response ) || is_wp_error( $response ) ) {
				return $this->default_zip_code();
			}

			$json = json_decode( $response['body'], true );

			if ( $json['error'] ) {
				return $this->default_zip_code();
			}

			return $json['postal'];
		}
	}
}
