<?php
/**
 * Custom WooCommerce Email class.
 * This is the main email sending class.
 *
 * @package NeverSettle\WooCommerce-Amazon-Fulfillment
 * @since 4.4.0
 */

defined( 'ABSPATH' ) || exit;

if ( ! class_exists( 'WC_Email' ) ) {
	return;
}

/**
 * Main email class for FBA emails.
 */
class NS_MCF_FBA_Order_Email extends WC_Email {

	/**
	 * Main constructor.
	 */
	public function __construct() {

		// set ID, this simply needs to be a unique name.
		$this->id             = 'ns_mcf_fba_order_email';
		$this->customer_email = true;

		// this is the title in WooCommerce Email settings.
		$this->title = 'MCF FBA Order Email';

		// this is the description in WooCommerce email settings.
		$this->description = 'MCF FBA Order notifications for patial or sent to fba orders.';

		// these define the locations of the templates that this email should use, we'll just use the new order template since this email is similar.
		$this->template_html  = 'emails/customer-sent-to-fba-order.php';
		$this->template_plain = 'emails/plain/customer-sent-to-fba-order.php';

		$this->placeholders = array(
			'{order_date}'   => '',
			'{order_number}' => '',
		);

		// Trigger on new fba orders.
		add_action( 'woocommerce_order_status_processing_to_sent-to-fba_notification', array( $this, 'trigger' ), 10, 2 );
		add_action( 'woocommerce_order_status_processing_to_part-to-fba_notification', array( $this, 'trigger' ), 10, 2 );

		// Call parent constructor to load any other defaults not explicity defined here.
		parent::__construct();
	}

	/**
	 * Get email subject.
	 *
	 * @return string
	 */
	public function get_default_subject() {
		return __( 'Your {site_title} order is being fulfilled!', 'ns-fba-for-woocommerce' );
	}

	/**
	 * Get email heading.
	 *
	 * @return string
	 */
	public function get_default_heading() {
		return __( 'Thank you for your order', 'ns-fba-for-woocommerce' );
	}

	/**
	 * Trigger the sending of this email.
	 *
	 * @param int            $order_id The order ID.
	 * @param WC_Order|false $order Order object.
	 */
	public function trigger( $order_id, $order = false ) {
		$this->setup_locale();

		if ( $order_id && ! is_a( $order, 'WC_Order' ) ) {
			$order = wc_get_order( $order_id );
		}

		if ( is_a( $order, 'WC_Order' ) ) {
			$this->object                         = $order;
			$this->recipient                      = $this->object->get_billing_email();
			$this->placeholders['{order_date}']   = wc_format_datetime( $this->object->get_date_created() );
			$this->placeholders['{order_number}'] = $this->object->get_order_number();
		}

		if ( $this->is_enabled() && $this->get_recipient() ) {
			$this->send( $this->get_recipient(), $this->get_subject(), $this->get_content(), $this->get_headers(), $this->get_attachments() );
		}

		$this->restore_locale();
	}

	/**
	 * Get content html.
	 *
	 * @return string
	 */
	public function get_content_html() {
		return wc_get_template_html(
			$this->template_html,
			array(
				'order'              => $this->object,
				'email_heading'      => $this->get_heading(),
				'additional_content' => $this->get_additional_content(),
				'sent_to_admin'      => false,
				'plain_text'         => false,
				'email'              => $this,
			)
		);
	}

	/**
	 * Get content plain.
	 *
	 * @return string
	 */
	public function get_content_plain() {
		return wc_get_template_html(
			$this->template_plain,
			array(
				'order'              => $this->object,
				'email_heading'      => $this->get_heading(),
				'additional_content' => $this->get_additional_content(),
				'sent_to_admin'      => false,
				'plain_text'         => true,
				'email'              => $this,
			)
		);
	}

	/**
	 * Default content to show below main email content.
	 *
	 * @since 3.7.0
	 * @return string
	 */
	public function get_default_additional_content() {
		return __( 'Thank you for your order.', 'ns-fba-for-woocommerce' );
	}


	/**
	 * Initialise Settings Form Fields
	 *
	 * @access public
	 * @return void
	 */
	public function init_form_fields() {
		// translators: %s: The available placeholders..
		$placeholder_text  = sprintf( __( 'Available placeholders: %s', 'ns-fba-for-woocommerce' ), '<code>' . esc_html( implode( '</code>, <code>', array_keys( $this->placeholders ) ) ) . '</code>' );
		$this->form_fields = array(
			'enabled'            => array(
				'title'   => _x( 'Enable/Disable', 'an email notification', 'ns-fba-for-woocommerce' ),
				'type'    => 'checkbox',
				'label'   => __( 'Enable this email notification', 'ns-fba-for-woocommerce' ),
				'default' => 'no',
			),
			'subject'            => array(
				'title'       => _x( 'Subject', 'of an email', 'ns-fba-for-woocommerce' ),
				'type'        => 'text',
				// translators: %s: default e-mail subject.
				'description' => sprintf( __( 'This controls the email subject line. Leave blank to use the default subject: %s.', 'ns-fba-for-woocommerce' ), '<code>' . $this->subject . '</code>' ),
				'placeholder' => $this->get_default_subject(),
				'default'     => '',
			),
			'heading'            => array(
				'title'       => _x( 'Email Heading', 'Name the setting that controls the main heading contained within the email notification', 'ns-fba-for-woocommerce' ),
				'type'        => 'text',
				// translators: %s: default e-mail heading.
				'description' => sprintf( __( 'This controls the main heading contained within the email notification. Leave blank to use the default heading: %s.', 'ns-fba-for-woocommerce' ), '<code>' . $this->heading . '</code>' ),
				'placeholder' => $this->get_default_heading(),
				'default'     => '',
			),
			'additional_content' => array(
				'title'       => __( 'Additional content', 'ns-fba-for-woocommerce' ),
				'description' => __( 'Text to appear below the main email content.', 'ns-fba-for-woocommerce' ) . ' ' . $placeholder_text,
				'css'         => 'width:400px; height: 75px;',
				'placeholder' => __( 'N/A', 'ns-fba-for-woocommerce' ),
				'type'        => 'textarea',
				'default'     => $this->get_default_additional_content(),
				'desc_tip'    => true,
			),
			'email_type'         => array(
				'title'       => _x( 'Email type', 'text, html or multipart', 'ns-fba-for-woocommerce' ),
				'type'        => 'select',
				'description' => __( 'Choose which format of email to send.', 'ns-fba-for-woocommerce' ),
				'default'     => 'html',
				'class'       => 'email_type',
				'options'     => array(
					'plain'     => _x( 'Plain text', 'email type', 'ns-fba-for-woocommerce' ),
					'html'      => _x( 'HTML', 'email type', 'ns-fba-for-woocommerce' ),
					'multipart' => _x( 'Multipart', 'email type', 'ns-fba-for-woocommerce' ),
				),
			),
		);
	}
}
